package tournament.test.configuration

import tournament.test.player.exceptions.AlreadyExistingPlayerException
import io.ktor.http.*
import io.ktor.server.application.*
import io.ktor.server.plugins.*
import io.ktor.server.plugins.statuspages.*
import io.ktor.server.response.*
import kotlinx.serialization.Serializable
import software.amazon.awssdk.services.dynamodb.model.DynamoDbException

private const val SOMETHING_WENT_WRONG = "Something went wrong"
private const val RESOURCE_NOT_FOUND = "Resource not found"

@Serializable
data class ErrorResponse(
    val message: String,
    val statusCode: Int,
)

fun Application.configureExceptionHandling() {
    install(StatusPages) {
        exception<Throwable> { call, cause ->
            val (errorStatusCode, errorResponse) = when (cause) {
                is NotFoundException -> HttpStatusCode.NotFound to ErrorResponse(
                    message = cause.message ?: RESOURCE_NOT_FOUND,
                    statusCode = HttpStatusCode.NotFound.value
                )

                is DynamoDbException -> HttpStatusCode.ServiceUnavailable to ErrorResponse(
                    message = cause.message ?: "$SOMETHING_WENT_WRONG with DynamoDb",
                    statusCode = HttpStatusCode.ServiceUnavailable.value,
                )

                is AlreadyExistingPlayerException -> HttpStatusCode.BadRequest to ErrorResponse(
                    message = cause.message ?: SOMETHING_WENT_WRONG,
                    statusCode = HttpStatusCode.BadRequest.value,
                )

                else -> throw cause
            }
            call.respond(errorStatusCode, errorResponse)
        }
    }
}