package tournament.test.player.repositories

import tournament.test.player.entities.Player
import tournament.test.player.entities.PlayerEntity
import tournament.test.player.entities.toPlayer
import tournament.test.player.entities.toPlayerEntity
import dev.andrewohara.dynamokt.DataClassTableSchema
import kotlinx.coroutines.future.await
import kotlinx.coroutines.reactive.asFlow
import software.amazon.awssdk.enhanced.dynamodb.DynamoDbEnhancedAsyncClient
import software.amazon.awssdk.enhanced.dynamodb.Expression
import software.amazon.awssdk.enhanced.dynamodb.Key
import software.amazon.awssdk.enhanced.dynamodb.model.ScanEnhancedRequest
import software.amazon.awssdk.services.dynamodb.model.AttributeValue

class PlayerRepositoryImpl(dynamoDbEnhancedClient: DynamoDbEnhancedAsyncClient) : PlayerRepository {
    private val tableName = PlayerEntity::class.simpleName
    private val tableSchema = DataClassTableSchema(PlayerEntity::class)
    private val table = dynamoDbEnhancedClient.table(tableName, tableSchema)

    override suspend fun create(player: Player): Player {
        table.putItem(player.toPlayerEntity()).await()
        return player
    }

    override suspend fun update(player: Player): Player {
        val updatedPlayer = table.updateItem(player.toPlayerEntity()).await()
        return updatedPlayer.toPlayer()
    }

    override suspend fun findByPseudo(pseudo: String): Player? {
        val foundPlayer = table.getItem(
            Key.builder().partitionValue(pseudo).build()
        ).await()
        return foundPlayer?.toPlayer()
    }

    override suspend fun getRank(player: Player): Int {
        val scan = ScanEnhancedRequest.builder()
            .filterExpression(
                Expression.builder()
                    .expression("pointsNumber > :pointsNumber")
                    .putExpressionValue(
                        ":pointsNumber", AttributeValue.builder().n(player.pointsNumber.toString()).build()
                    ).build()
            ).build()
        return buildList {
            table.scan(scan).asFlow().collect { it.items().stream().forEach { t -> add(t) } }
        }.size + 1
    }

    override suspend fun findAll(): List<Player> {
        return buildList {
            table.scan().asFlow().collect { it.items().stream().forEach { item -> add(item.toPlayer()) } }
        }
    }


    override suspend fun deleteAll() {
        table.deleteTable()
        table.createTable().await()

    }
}