package tournament.test.player.services

import tournament.test.player.dtos.PlayerCreationDTO
import tournament.test.player.dtos.PlayerInfoDTO
import tournament.test.player.dtos.PlayerUpdateDTO
import tournament.test.player.dtos.toPlayer
import tournament.test.player.dtos.toPlayerInfoDTO
import tournament.test.player.entities.Player
import tournament.test.player.exceptions.AlreadyExistingPlayerException
import tournament.test.player.repositories.PlayerRepository
import io.ktor.server.application.*
import io.ktor.server.plugins.*
import org.slf4j.LoggerFactory

class PlayerServiceImpl(private val playerRepository: PlayerRepository) : PlayerService {
    private val logger = LoggerFactory.getLogger(Application::class.java)

    override suspend fun createNewPlayer(playerCreationDTO: PlayerCreationDTO): Player {
        logger.info("Creating new player")
        if (playerRepository.findByPseudo(playerCreationDTO.pseudo) != null) {
            throw AlreadyExistingPlayerException(playerCreationDTO.pseudo)
        }
        val created = playerRepository.create(playerCreationDTO.toPlayer())
        logger.info("Player ${created.pseudo} successfully created")
        return created
    }

    override suspend fun updatePlayer(playerUpdateDTO: PlayerUpdateDTO): Player {
        logger.info("Updating player ${playerUpdateDTO.pseudo}")
        getPlayerByPseudo(playerUpdateDTO.pseudo)
        val updated = playerRepository.update(playerUpdateDTO.toPlayer())
        logger.info("Player ${updated.pseudo} successfully updated")
        return updated
    }

    override suspend fun getPlayerByPseudo(pseudo: String): Player {
        return playerRepository.findByPseudo(pseudo) ?: throw NotFoundException("Player $pseudo not found")
    }

    override suspend fun getPlayerInfoByPseudo(pseudo: String): PlayerInfoDTO {
        logger.info("Gathering infos for $pseudo")
        val player = getPlayerByPseudo(pseudo)
        val rank = playerRepository.getRank(player)
        return player.toPlayerInfoDTO(rank)
    }

    override suspend fun getPlayersRanked(): List<PlayerInfoDTO> {
        val allPlayers = playerRepository.findAll()
        var currentRank = 1
        var previousPoints = 0
        logger.info("Sorting all player by their rank")
        val sorted = allPlayers.sortedByDescending { it.pointsNumber }.mapIndexed { index, player ->
            if (previousPoints != player.pointsNumber) {
                currentRank = index + 1
            }
            previousPoints = player.pointsNumber
            PlayerInfoDTO(
                pseudo = player.pseudo,
                pointsNumber = player.pointsNumber,
                ranking = currentRank,
            )
        }
        return sorted

    }

    override suspend fun deleteAllPlayers() {
        logger.info("Deleting all players from database")
        playerRepository.deleteAll()
        logger.info("Players successfully deleted")
    }
}