package player

import tournament.test.player.dtos.PlayerCreationDTO
import tournament.test.player.dtos.PlayerUpdateDTO
import tournament.test.player.entities.Player
import tournament.test.player.exceptions.AlreadyExistingPlayerException
import tournament.test.player.repositories.PlayerRepository
import tournament.test.player.services.PlayerServiceImpl
import io.ktor.server.plugins.*
import io.mockk.coEvery
import io.mockk.coVerify
import io.mockk.just
import io.mockk.mockk
import io.mockk.runs
import kotlinx.coroutines.runBlocking
import org.assertj.core.api.Assertions.assertThat
import org.junit.Assert.assertThrows
import org.junit.Test

class PlayerServiceTest {
    private val playerRepository: PlayerRepository = mockk()

    private val playerService: PlayerServiceImpl = PlayerServiceImpl(playerRepository)
    private val john = "John"
    private val doe = "Doe"
    private val player1 = Player(pseudo = john)
    private val playerToUpdate = Player(pseudo = john, pointsNumber = 10)
    private val player2 = Player(pseudo = doe, pointsNumber = 20)

    @Test
    fun `when creating player, should call the repository once`() {
        coEvery { playerRepository.create(player1) } returns player1
        coEvery { playerRepository.findByPseudo(john) } returns null
        runBlocking { playerService.createNewPlayer(PlayerCreationDTO(pseudo = john)) }
        coVerify(exactly = 1) { playerRepository.create(player1) }
    }

    @Test
    fun `when creating player that exists, should throw on existing player`() {
        coEvery { playerRepository.findByPseudo(john) } returns player1
        assertThrows(AlreadyExistingPlayerException::class.java) {
            runBlocking { playerService.createNewPlayer(PlayerCreationDTO(pseudo = john)) }
        }
    }

    @Test
    fun `when updating player, should call the repository once`() {
        coEvery { playerRepository.findByPseudo(pseudo = john) } returns player1
        coEvery { playerRepository.update(playerToUpdate) } returns playerToUpdate
        runBlocking { playerService.updatePlayer(PlayerUpdateDTO(pseudo = john, pointsNumber = 10)) }
        coVerify(exactly = 1) { playerRepository.update(playerToUpdate) }
    }

    @Test
    fun `when updating player that doesn't exist, should throw a not found player`() {
        coEvery { playerRepository.findByPseudo(pseudo = john) } returns null
        assertThrows(NotFoundException::class.java) {
            runBlocking { playerService.updatePlayer(PlayerUpdateDTO(pseudo = john, pointsNumber = 10)) }
        }
    }

    @Test
    fun `when finding user by pseudo, should call the repository once`() {
        coEvery { playerRepository.findByPseudo(pseudo = john) } returns player1
        runBlocking { playerService.getPlayerByPseudo(pseudo = john) }
        coVerify(exactly = 1) { playerRepository.findByPseudo(john) }
    }

    @Test
    fun `when finding by pseudo without match, should throw a not found player`() {
        coEvery { playerRepository.findByPseudo(pseudo = john) } returns null
        assertThrows(NotFoundException::class.java) {
            runBlocking { playerService.getPlayerByPseudo(pseudo = john) }
        }
    }

    @Test
    fun `when getting player info, should return player with his rank`() {
        coEvery { playerRepository.findByPseudo(pseudo = john) } returns player1
        coEvery { playerRepository.getRank(player1) } returns 2
        val result = runBlocking { playerService.getPlayerInfoByPseudo(pseudo = john) }
        assertThat(result.pseudo).isEqualTo(john)
        assertThat(result.ranking).isEqualTo(2)
    }

    @Test
    fun `when getting player info on non-existing player, should throw a not found player exception`() {
        coEvery { playerRepository.findByPseudo(pseudo = john) } returns null
        coEvery { playerRepository.getRank(player1) } returns 2
        assertThrows(NotFoundException::class.java) {
            runBlocking { playerService.getPlayerInfoByPseudo(pseudo = john) }
        }
    }

    @Test
    fun `when getting all players ranking, should return players sorted by rank`() {
        coEvery { playerRepository.findAll() } returns listOf(player1, player2)
        val result = runBlocking { playerService.getPlayersRanked() }
        assertThat(result.first().ranking).isEqualTo(1)
        assertThat(result.last().ranking).isEqualTo(2)
    }

    @Test
    fun `when getting all players with two players with the same points number, should return the same rank`() {
        val player2WithSamePoints = player2.copy(pointsNumber = 30)
        val player3WithSamePoints = Player(pseudo = "Charlie", pointsNumber = 30)
        coEvery { playerRepository.findAll() } returns listOf(player1, player2WithSamePoints, player3WithSamePoints)
        val result = runBlocking { playerService.getPlayersRanked() }
        assertThat(result.first().ranking).isEqualTo(result[1].ranking)
        assertThat(result.last().ranking).isNotEqualTo(2)
    }

    @Test
    fun `when deleting all players, should call the repository once`() {
        coEvery { playerRepository.deleteAll() } just runs
        runBlocking { playerService.deleteAllPlayers() }
        coVerify(exactly = 1) { playerRepository.deleteAll() }
    }
}